#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
:package: lazydog
:synopsis: File system user-level events monitoring.
:author: Clément Warneys <clement.warneys@gmail.com>

This is the main package of the **lazydog** library. It relies on another 
sub-package **revised_watchdog** which is a modified version of **watchdog**
package.

As a summary, **watchdog** is a "*python API and shell utilities to monitor file 
system events*". As such, **watchdog** is monitoring and emitting every tiny 
local event on the file system, which often means 5 or more watchdog events per
user event. For example, when a user is creating a new file, you will get 1 creation
event, multiple modification events (some of them for the content modification, others 
for metadata mosification), and 1 or more modification event for the directory of 
the file.

The goal of **lazydog** is to emit only 1 event per user event. This kind of event 
will sometimes be call **high-level event**, compared to **low-level event**
which are emitted by the **watchdog** API. To do so, **lazydog** is waiting a little 
amount of time in order to correlate different **watchdog** events between them, and to 
aggregate them when related. This mechanism results in some delays between the user action
and the event emission. The total delay depends on the **watchdog** observer class. For 
example, if you use an :class:`InotifyObserver` observer, you only need a 2-seconds delay. But if you use
a more basic observer as the :class:`PollingObserver` observer (which is more compatible between different 
system), then you need a greater delay such as 10-seconds.

The **lazydog** package contains the following modules:

* :py:mod:`~lazydog.lazydog` is a sample module that show how to use the package, \
by logging the high-level events in the console. The main function of this module is  \
called when calling ``$ lazidog`` in the console.
* :py:mod:`~lazydog.handlers` is the main module of the library with the aggregation algorithms. 
* :py:mod:`~lazydog.events` defines the high-level lazydog events, based on the \
low-level watchdog ones, which are now aggregable and also convertible \
to copy or move events.
* :py:mod:`~lazydog.queues` bufferizes lazydog events pending for a possible \
aggregation with other simultaneous events.
* :py:mod:`~lazydog.states` keeps track of the current state of the watched local directory. \
The idea is to save computational time, avoiding recomputing file hashes \
or getting size and time of each watched files (depending on the requested \
method), thus facilitating identification of copy events.
* :py:mod:`~lazydog.dropbox_content_hasher` is the default hash function to get a hash of a file. \
Based on the hash function of the Dropbox API.


lazydog.lazydog
===============

.. automodule:: lazydog.lazydog
   :members:

lazydog.handlers
================

.. automodule:: lazydog.handlers
   :members:

lazydog.events
==============

.. automodule:: lazydog.events
   :members:

lazydog.queues
==============

.. automodule:: lazydog.queues
   :members:

lazydog.states
==============

.. automodule:: lazydog.states
   :members:

lazydog.dropbox_content_hasher
==============================

.. automodule:: lazydog.dropbox_content_hasher
   :members:

"""

__version__ = '0.1.1'




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: lazydog.states
:synopsis: Keeps track of the current state of the watched local directory. \
The idea is to save computational time, avoiding recomputing file hashes \
or getting size and modification time of each watched files (depending on the requested \
method), thus accelerating identification of copy events.
:author: Clément Warneys <clement.warneys@gmail.com>

"""

import os
import logging

from lazydog.dropbox_content_hasher import default_hash_function

class DualAccessMemory():
    """
    Helper class, used by :py:class:`LocalState`. Sort of double-entry dictionary.
    When you save one tuple {key, value}, you can then access it both way:

    * either from the key, using :py:meth:`get`, or using accessor ``object[key]``
    * or from value, using :py:meth:`get_by_value`. In this case, you will get a set \
    of all the corresponding keys that references to this specific value.

    To register a new key, you can either use :py:meth:`save` method, 
    or the accessor ``object[key] = value``.

    Finally you can check if a key is existing using the accessor ``key in object``.

    .. note: This dual-access dictionary has been specifically designed for
        handling `path` key (for example '/dir/file.txt'), and associate it
        to a file value, which can be either file size, file modification time, 
        or file hash...

        Thus the :py:class:`DualAccessMemory` class contains specifics methods
        allowing to :py:meth:`move` and :py:meth:`delete` keys recursively, 
        according to the way file-system paths are recursively moved or deleted.
    """
    
    def __init__(self):
        # self.memories is a dictionary with unique keys and possible same values
        # note that this class has been designed for keys that should be a string containing a path
        self.memories = {}
        self.memories.clear()
        
        # self.dual_memories is a dictionary where the keys are each possible value of self.memories, 
        # and the values are the lists of the associated keys.
        self.dual_memories = {}
        self.dual_memories.clear()

    def get(self, key):
        """
        Returns the value corresponding to the key in parameter, same behaviour as a dictionary. 
        ``None`` if key is unknowned. You can also access it with ``object[key]``.
        """
        return self.memories.get(key, None)
    
    def get_by_value(self, value) -> set:
        """
        Returns a set of key corresponding to the value in parameter. 
        Empty ``set()`` if value is not referenced.
        """
        return self.dual_memories.get(value, set())

    def __getitem__(self, key):
        return self.get(key)
    
    def __setitem__(self, key, value):
        return self.save(key, value)

    def __contains__(self, key):
        return key in self.memories
        
    def _get_children(self, key:str):
        """
        Private inner helper method. Considering the :py:class:`DualAccessMemory` 
        has been designed to handle path key, this method returns a list of every 
        children paths under the key path (including the key path itself).
        """
        complete_key = key if key.endswith('/') else key + '/'
        children = [x for x in self.memories if x.startswith(complete_key)]
        if key in self.memories:
            children = children + [key]
        return children

    def save(self, key:str, value):
        """
        Registers the tuple {key, value} in order it is easily accessible both way.
        If key already exists with another value, the value is first removed, 
        before registering the new one.
        """
        if key in self:
            self.dual_memories[self.memories[key]].discard(key)
        self.memories[key] = value
        if self.dual_memories.get(self.memories[key]) is None:
            self.dual_memories[self.memories[key]] = set()
        self.dual_memories[self.memories[key]].update([key]) 
    
    # Delete key recursively
    def delete(self, delete_key:str):
        """
        Considering the :py:class:`DualAccessMemory` 
        has been designed to handle path key, this method not only deletes the
        ``delete_key`` in parameter, but it also deletes every children keys corresponding to the children paths 
        of the parameter path ``delete_key``.
        """
        for key in self._get_children(delete_key):
            if self.dual_memories.get(self.memories[key]) is not None:
                self.dual_memories[self.memories[key]].discard(key) 
            self.memories.pop(key)
        
    # Move key recursively
    def move(self, src_key:str, dst_key:str):
        """
        Considering the :py:class:`DualAccessMemory` 
        has been designed to handle path key, this method not only moves the
        ``src_key`` in parameter to ``dst_key`` key, but it also moves every 
        children keys corresponding to the children paths 
        of the parameter path ``src_key`` to the related children path under 
        the parameter path ``dst_key``.
        """
        for old_key in self._get_children(src_key):
            new_key = old_key.replace(src_key, dst_key, 1)
            if new_key in self:
                self.dual_memories[self.memories[new_key]].discard(new_key)
            if self.dual_memories.get(self.memories[old_key]) is not None:
                self.dual_memories[self.memories[old_key]].discard(old_key) 
                self.dual_memories[self.memories[old_key]].update([new_key]) 
            self.memories[new_key] = self.memories.pop(old_key)
            
            
    
    
class LocalState():
    """
    Keeps track of the current state of the watched local directory, by listing 
    every sub-files and sub-directories, and associating each of them with their
    size, modification time, and hash values.

    When managing large directory, it can become very long to retrieves this 
    information. But we need it very fast in order to be able to correlate 
    `Created` event into `Copied` ones. Indeed, for this kind of correlation, 
    we need to rapidly find every other file or folder that are having the 
    same characteristics (that will then be eligible to be the source file 
    or folder).

    :py:class:`LocalState` is keeping tracks of files with two 
    :py:class:`DualAccessMemory` objects. The first one keeping tracks of
    couple ``(size, modification time)``, and the second one of single 
    ``hash`` value.

    Hash values are computed depending on a default hashing function. This
    default method is based on the Dropbox hashing algorithm, but you can
    define your own one. You only have to respect the same parameter and return.
    See :py:meth:`_default_hashing_function` method to see the needed parameters 
    names and types and the return type.

    In order to accelerate the initialization of :py:class:`LocalState` when 
    watching large diectory, you can initialize it with pre-computed initializing
    values of your own (that you have to know in the first place, for example by
    keeping track of them in a hard backup file, or if you already have to compute them
    in other place of your application, no need that the hash values have to 
    be computed again... just send them at the initialization). Please looke 
    at the ``custom_intializing_values`` parameter for more information.

    .. note: Once initialized, the :py:class:`LocalState` object needs to 
        be updated by external objects using it. Local state is not 
        automatically updating its own states, but sometimes when it detects 
        that some paths do not exist anymore (then it deletes them, for 
        cleaning purpose).

    :param absolute_root_folder:
        Absolute path of the folder you need to keep track of. Note that
        ever sub-file and sub-folder will then be referenced with
        relative paths.
    :type absolute_root_folder:
        str
    :param custom_hash_function:
        *Optional*. Default value is :py:meth:`_default_hashing_function` is used, 
        which is based on the Dropbox hashing algorithm. But you can also
        provides your own hashing function, as long as your respect the format
        of the default one.
    :type custom_hash_function:
        function
    :param custom_intializing_values:
        *Optional*. If not provided or ``None``, all sub-folders will be browsed
        at initialization, and for each file and folder, the file size, file 
        modification time and file hash will be retrieves and computed (this operation
        can take a long time, depending on the number and size of the files, and on 
        the hashing function). To accelerate this initialization process, you can provide
        __init__ method with pre-computed initializing values under a dictionary format with 
        ``key=file_path`` and ``value=[file_hash, file_size, file_time]``.
        You do not need to know the exact content of the main directory at the initialization, 
        and if you later notice unexpected modifications compared to the initial values you sent, 
        you can still correct each of them using the :py:meth:`save` method.
    :type custom_intializing_values:
        dict

    :returns: 
        An initialized object representing local state of the aimed folder.
    :rtype: 
        :py:class:`LocalState`

    """
    
    DEFAULT_DIRECTORY_VALUE = 'DIR'
    """
    Default hash value for directory (since directory are not hashed, 
    and that we want to reserve ``None`` value to non existing directories).
    """
    
    # Default method to get the Hash of the file with the supplied file name.
    @staticmethod
    def _default_hashing_function(absolute_path:str): 
        """
        Hash values are computed depending on a default hashing function. This
        default method is based on the Dropbox hashing algorithm. For information, 
        if the ``absolute_path`` is a directory, the directory hash will always be 
        the same, here :py:attr:`DEFAULT_DIRECTORY_VALUE`.

        :param absolute_path:
            Absolute path of the file or folder.
        :type absolute_path:
            str
       
        :returns: 
            The hash of the file or folder in parameter
        :rtype: 
            str
        """
        return default_hash_function(absolute_path, LocalState.DEFAULT_DIRECTORY_VALUE)
    
    # Following 3 methods can be used in other classes
    def absolute_local_path(self, relative_path:str) -> str:
        """
        Computes the absolute local path from a relative one.

        .. note: Since the instance of :py:class:`LocalState` contains the reference of 
            the watched directory, this method can be used from any other place where 
            the absolute local path is needed, as a generic helper method.

        :param relative_path:
            Relative local path of the file or folder.
        :type relative_path:
            str
        :returns: 
            Absolute local path of the same file or folder
        :rtype: 
            str
        """
        if relative_path.startswith('/'):
            relative_path = relative_path[1:]
        return os.path.join(self.absolute_root_folder, relative_path)
    
    def relative_local_path(self, absolute_path:str) -> str:
        """
        Same as :py:meth:`absolute_local_path`, but opposite.
        """
        absolute_path = os.path.normpath(absolute_path)
        return '/' + os.path.relpath(absolute_path, self.absolute_root_folder)
    
    def hash_function(self, *args, **kwargs):
        return self._hash_function(*args, **kwargs)
    
    def __init__(self, absolute_root_folder, custom_hash_function=None, custom_intializing_values:dict=None):
        # keep absolute root folder
        self.absolute_root_folder = absolute_root_folder
        
        # keep hash function
        self._hash_function = custom_hash_function if custom_hash_function is not None else LocalState._default_hashing_function
        
        # self.hashes is a dual access dictionary 
        # self.hashes.get(key) with key=file_path returns the value=file_hash
        # self.hashes.get_by_value(value) with value=file_hash returns a set of paths
        self.hashes = DualAccessMemory()
        
        # self.sizetimes is a dual access dictionary 
        # self.sizetimes.get(key) with key=file_path returns the value=tuple(file_size, file_mtime)
        # self.sizetimes.get_by_value(value) with value=tuple(file_size, file_mtime) returns a set of paths
        self.sizetimes = DualAccessMemory()
        
        # Initializing values
        if custom_intializing_values is not None:
            # custom_intializing_values should be a dict with:
            # - key=file_path
            # - value=list(file_hash, file_size, file_time)
            for k, v in custom_intializing_values.items():
                if os.path.exists(self.absolute_local_path(k)):
                    self.save(k, v[0], v[1], v[2])
                    logging.debug('Initial indexing (provided) ' + k + ' - ' + v[0] + ' - ' + str((v[1], v[2])))
        else:
            
            # Default initializing
            for root, dirs, files in os.walk(self.absolute_root_folder):  
                for i in dirs + files:
                    relative_path = self.relative_local_path(os.path.join(root, i))
                    self.get_hash(relative_path)
                    self.get_sizetime(relative_path)
                    logging.debug('Initial indexing (computed) ' + relative_path + ' - ' + self.get_hash(relative_path) + ' - ' + str(self.get_sizetime(relative_path)))
    

    
    def get_hash(self, key:str, compute_if_none:bool=True) -> str:
        """
        Gets the ``file_hash`` value of the file at the ``key`` relative path. If the file is unknown
        (and so the hash value is not yet computed), by default the hash value will
        be computed. This behaviour can be cancelled using ``compute_if_none`` parameter.

        :param key:
            Relative local path of the file or folder.
        :type key:
            str
        :param compute_if_none:
            *Optional*. ``True`` by default, which means that if the file is unknown
            (and so it is for the hash value), the hash value will be computed. Use
            ``False`` if you want to cancel this bahaviour, so the returned value
            will be ``None``.
        :type compute_if_none:
            boolean
        :returns: 
            File or directory hash value, if path exists, else ``None``.
        :rtype: 
            str
        """
        if key not in self.hashes and compute_if_none:
            self.hashes[key] = self.hash_function(self.absolute_local_path(key))
        return self.hashes[key]
        
    def get_files_by_hash_key(self, hash_key:str) -> set:
        """
        Returns a set of every file or directory paths for which the 
        hash value corresponds to the ``hash_key`` parameter.
        """
        file_paths = self.hashes.get_by_value(hash_key)
        return self._check_for_deleted_paths(file_paths)

    def get_sizetime(self, key:str, compute_if_none:bool=True):
        """
        Gets the couple ``(file_size, file_modification_time)`` value of the file 
        at the ``key`` relative path. Same behaviour than :py:meth:`get_hash` method.

        :param key:
            Relative local path of the file or folder.
        :type key:
            str
        :param compute_if_none:
            *Optional*. ``True`` by default, which means that if the file is unknown
            (and so it is for the file size and modification time value), the values 
            will be computed. Use ``False`` if you want to cancel this bahaviour, so the 
            returned value will be ``None``.
        :type compute_if_none:
            boolean
        :returns: 
            File or directory couple (file_size, file_modification_time) value, 
            if path exists, else ``None``.
        :rtype: 
            str
        """
        if key not in self.sizetimes and compute_if_none:
            if os.path.isdir(self.absolute_local_path(key)):
                self.sizetimes[key] = (LocalState.DEFAULT_DIRECTORY_VALUE, 
                                       LocalState.DEFAULT_DIRECTORY_VALUE)
            elif os.path.exists(self.absolute_local_path(key)):
                self.sizetimes[key] = (os.path.getsize(self.absolute_local_path(key)), 
                                       round(os.path.getmtime(self.absolute_local_path(key)), 3))
        return self.sizetimes[key]
        
    def get_files_by_sizetime_key(self, sizetime_key) -> set:
        """
        Returns a set of every file or directory paths for which the 
        couple (file_size, file_modification_time) value corresponds 
        to the ``sizetime_key`` parameter.
        """
        file_paths = self.sizetimes.get_by_value(sizetime_key)
        return self._check_for_deleted_paths(file_paths)
    
    def _check_for_deleted_paths(self, paths:set):
        deleted_paths = [x for x in paths if not os.path.exists(self.absolute_local_path(x))]
        for dp in deleted_paths:
            self.hashes.delete(dp)
            self.sizetimes.delete(dp)
        return paths - set(deleted_paths)
    
    def save(self, key:str, file_hash, file_size, file_mtime):
        """
        Allows an external object to add a new file or folder reference to the local state object, 
        by giving already computed hash, size and modification time values. Note that the values
        will not be neither checked nor recomputed.

        If you prefer that the :py:class:`LocalState` class computes these values itself, and add
        the file or folder reference, you can just call the :py:meth:`get_hash` or 
        :py:meth:`get_sizetime` method. Note that then the :py:class:`LocalState` object just compute
        the needed values: it can compute the hash value without having any reference 
        in its sizetime dictionary. These one will only be computed when calling the 
        related method.

        :param key:
            Relative local path of the file or folder.
        :type key:
            str
        :param file_hash:
            File hash value of the file or folder.
        :type file_hash:
            str
        :param file_size:
            File size value of the file or folder. For information
            the size is computed with :py:meth:`os.path.getsize()`
            method, so the size is the number of bytes of the file.
        :type file_size:
            int
        :param file_mtime:
            File modification time value of the file or folder. For information
            the modification time is computed with :py:meth:`os.path.getmtime()`
            method, rounded to the third decimal, so the time is a number giving 
            the number of seconds since the epoch, precise at the millisecond.
        :type file_mtime:
            int
        :returns: 
            ``None``
        """
        if os.path.isdir(self.absolute_local_path(key)):
            file_hash = LocalState.DEFAULT_DIRECTORY_VALUE
            file_size = LocalState.DEFAULT_DIRECTORY_VALUE
            file_mtime = LocalState.DEFAULT_DIRECTORY_VALUE
        self.hashes[key] = file_hash
        self.sizetimes[key] = (file_size, file_mtime)
    
    def delete(self, delete_key:str):
        """
        Deletes key recursively. This method can be called internally 
        when detecting a file or folder does not exists anymore, or 
        by an external objects, that do not need to keep track of 
        this path anymore.
        """
        self.hashes.delete(delete_key)
        self.sizetimes.delete(delete_key)
    
    def move(self, src_key:str, dst_key:str):
        """
        Moves key recursively. This method can be called by an 
        external object, when you know a file or folder has been moved
        and that you want to keep the already computed values in 
        reference, without recomputing them all.
        """
        self.hashes.move(src_key, dst_key)
        self.sizetimes.move(src_key, dst_key)
            
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: lazydog.lazidog
:synopsis: An sample module that show how to use the package, \
by logging the high-level lazydog events in the console. \
The main function of this module is executed by calling \
``$ lazidog`` in the system console.
:author: Clément Warneys <clement.warneys@gmail.com>

Please read the source code for more information. Below is an example on 
how to initialize the high-level lazydog event handler, and log every 
new event in the console (using logging module). The watched directory 
is the current one (using ``os.getcwd()``).

.. code-block:: python 
   :linenos:

    import logging
    import os

    from lazydog.handlers import HighlevelEventHandler

    # LOG    
    # create logger 
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    # create console handler with a higher log level
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(console_handler)

    # INITIALIZE 
    # get dir in parameter else current dir
    watched_dir = directory if len(directory) > 1 else os.getcwd()
    # initializing a new HighlevelEventHandler
    highlevel_handler = HighlevelEventHandler.get_instance(watched_dir)
    # starting it (since it is a thread)
    highlevel_handler.start()
    # log first message
    logging.info('LISTENING EVENTS IN DIR: \'%s\'' % watched_dir)
        
    # OPERATING
    try:
        while True:

            # The following loop check every 1 second if any new event.
            time.sleep(1)
            local_events = highlevel_handler.get_available_events()
            
            # If any, it logs it directly in the console.
            for e in local_events:
                logging.info(e)

        # Keyboard <CTRL+C> interrupts the loop 
        except KeyboardInterrupt:   
            highlevel_handler.stop()


"""

import time
import logging
import os

from lazydog.handlers import HighlevelEventHandler



# this function shows an example of how to use Lazydog.
def logging_in_console(directory:str=''):
    if directory == None:
        directory = ''
    
    # LOG    
    # create logger 
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    # create console handler with a higher log level
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(console_handler)
    
    # INITIALIZE 
    # get dir in parameter else current dir
    watched_dir = directory if len(directory) > 1 else os.getcwd()
    # initializing a new HighlevelEventHandler
    highlevel_handler = HighlevelEventHandler.get_instance(watched_dir)
    # starting it (since it is a thread)
    highlevel_handler.start()
    # log first message
    logging.info('LISTENING EVENTS IN DIR: \'%s\'' % watched_dir)

    try:
        while True:

            # The following loop check every 1 second if any new event.
            time.sleep(1)
            local_events = highlevel_handler.get_available_events()

            # If any, it logs it directly in the console.
            for e in local_events:
                logging.info(e)
                
    # Keyboard <CTRL+C> interrupts the loop 
    except KeyboardInterrupt:   
        highlevel_handler.stop()
    




# this function is the one executed when calling '$ lazydog' in the system console.
def main():
    
    import sys
    directory_to_watch = sys.argv[1] if len(sys.argv) > 1 else None
    logging_in_console(directory_to_watch)
    
    
    
if __name__ == "__main__":
    main()
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
# Copyright 2014 Thomas Amland <thomas.amland@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: revised_watchdog.observers.inotify_c
:author: Thomas Amland <thomas.amland@gmail.com>
:author: Clément Warneys <clement.warneys@gmail.com>

This module is overloading the original :py:mod:`watchdog.observers.inotify_buffer` module 
by revising and completing it. Please read original **watchdog** project 
documentation for more information: https://github.com/gorakhargosh/watchdog

The main change is in the :py:class:`InotifyBuffer` class, whose :py:meth:`InotifyBuffer.__init__`
method now uses revised watchdog :py:class:`~lazydog.revised_watchdog.observers.inotify_c.Inotify` class.

"""

import logging
from watchdog.utils import BaseThread
from watchdog.utils.delayed_queue import DelayedQueue
from watchdog.observers.inotify_buffer import InotifyBuffer
from lazydog.revised_watchdog.observers.inotify_c import Inotify 

logger = logging.getLogger(__name__)

class InotifyBuffer(InotifyBuffer):
    """
    A wrapper for `Inotify` that holds events for `delay` seconds. During
    this time, ``IN_MOVED_FROM`` and ``IN_MOVED_TO`` events are paired.

    Please note that his class remains unmodified 
    in revised_watchdog package. Only the :py:meth:`__init__` method is overrided 
    in order it uses the new definition of :py:class:`~lazydog.revised_watchdog.observers.inotify_c.Inotify` class.
    """


    def __init__(self, path, recursive=False):
        BaseThread.__init__(self)
        self._queue = DelayedQueue(self.delay)
        self._inotify = Inotify(path, recursive)
        self.start()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: lazydog.handlers
:synopsis: Main module of lazydog including the aggregation logics.
:author: Clément Warneys <clement.warneys@gmail.com>

"""


import os
import time
import datetime
import threading

from lazydog.states import LocalState
from lazydog.events import LazydogEvent
from lazydog.queues import DatedlocaleventQueue

from lazydog.revised_watchdog.observers.inotify import InotifyEmitter, InotifyObserver
from watchdog.observers.polling import PollingObserver




class HighlevelEventHandler(threading.Thread):
    """
    Post-treats the low level events to suggest only high-level ones. 

    To do so, a high-level handler needs a :py:class:`~lazydog.queues.DatedlocaleventQueue`
    (an event queue containing the last lazydog events and inherited from 
    :py:class:`~revised_watchdog.events.FileSystemEventHandler` so it is compatible
    with watchdog observers), that is already populated by a low-level watchdog observer, 
    for example :py:class:`~revised_watchdog.observers.inotify.InotifyObserver`, that retrieves 
    the low-level file-system events.

    The simplest way to instanciate :py:class:`~lazydog.handlers.HighlevelEventHandler`
    is to use the :py:meth:`get_instance` method. In this case, you only need to specify the 
    directory to watch ``watched_dir``. Two other optional parameters ``hashing_function``
    and ``custom_intializing_values`` respectively allow to use custom hashing function 
    (which will be use to compute the hashs of each file, in order to correlate copy events)
    and to accelerate the initialization phase (by providing already computed hash values of the 
    current files in the watched directory, thus avoiding to compute them at the start). Please see 
    the related methods documentation for more information.

    This class inherits from :py:class:`threading.Thread` so it works autonomously. It can be started 
    :py:meth:`start` method (from `Thread` module, and a stopping order can be send with :py:meth:`stop` 
    inner method.


    :param lowlevel_event_queue:
        An event queue containing the last lazydog events. Note that the provided queue
        shall be already associated with a low-level watchdog observer that retrieves
        the low-level file-system events (in order to fill the queue).
    :type lowlevel_event_queue:
        :py:class:`~lazydog.queues.DatedlocaleventQueue`
    :param local_states:
        The reference state of the local files in the watched directory. This state will be 
        dynamically updated by the handler, depending on the low-level events. This state 
        contains also the path of the watched directory.
    :type local_states:
        :py:class:`~lazydog.states.LocalState`
    :returns: 
        A non-running high-level lazydog events handler.
    :rtype: 
        :py:class:`~lazydog.handlers.HighlevelEventHandler`
    """

    POSTTREATMENT_TIME_LIMIT = datetime.timedelta(seconds=2)
    """
    If neither new low-level events nor high-level post-treatments appends
    during this 2-seconds delay, the current events in the queue are ready to be
    emitted the listener, when using :py:meth:`get_available_events` method.
    """

    CREATE_EVENT_TIME_LIMIT_FOR_EMPTY_FILES = datetime.timedelta(minutes=15) # older behaviour...
    CREATE_EVENT_TIME_LIMIT_FOR_EMPTY_FILES = datetime.timedelta(seconds=2)
    """
    *Deprecated*. At the beginning of the project, empty file creation was more delayed before 
    the handler emits them. Because empty file are often created and then 
    rapidly renamed and modified... The idea was to limit the number of high-level
    events that were being sent. But this specific behaviour could generate unwanted 
    problems for the third-application using this library.
    """
    
    @classmethod
    def get_instance(cls, watched_dir:str, hashing_function=None, custom_intializing_values=None):
        """
        This method provides you with the  simplest way to instanciate 
        :py:class:`~lazydog.handlers.HighlevelEventHandler`. You only need to specify the 
        directory to watch ``watched_dir``. Two other optional parameters ``hashing_function``
        and ``custom_intializing_values`` respectively allow to use custom hashing function 
        (which will be use to compute the hashs of each file, in order to correlate copy events)
        and to accelerate the initialization phase (by providing already computed hash values of the 
        current files in the watched directory, thus avoiding to compute them at the start). 

        :param watched_dir:
            The path you want to watch.
        :type watched_dir:
            str
        :param hashing_function:
            Custom hashing function that will be use to compute the hashs of each file, 
            in order the handler is able to correlate copy events. The function shall be defined  
            with the same parameters and retrun format than 
            :py:func:`~lazydog.dropbox_content_hasher.default_hash_function`.
        :type hashing_function:
            function
        :param custom_intializing_values:
            Providing custom intializing values accelerate the initialization phase 
            by providing already computed hash values of all the files currently in 
            the watched directory. The provided dictionary shall cover all the local
            files and directory because hash values will not be computed for 
            missing files. If some reference were missing in the provided dictionary, 
            they can be completed later using :py:meth:`save_locals` 
            method. For more information about the structure of this parameter, 
            please see the documentation of :py:class:`~lazydog.states.LocalState`.
        :type custom_intializing_values:
            :py:class:`~lazydog.states.LocalState`
        :returns: 
            An already running high-level lazydog events handler.
        :rtype: 
            :py:class:`~lazydog.handlers.HighlevelEventHandler`
        """
        local_files = LocalState(watched_dir, hashing_function, custom_intializing_values)
        
        dated_event_queue = DatedlocaleventQueue(local_files)
        observer = InotifyObserver() # generate_full_events=False) # With reviewed Inotify 
        observer.schedule(dated_event_queue, watched_dir, recursive=True)
        observer.name = 'Local Inotify observer'
        observer.start()
        
        return cls(dated_event_queue, local_files)
    
    
    def __init__(self, lowlevel_event_queue:DatedlocaleventQueue, local_states:LocalState):
        self.events_list = []
        self.events_list.clear()
        self.lowlevel_event_queue = lowlevel_event_queue
        self.local_states = local_states
        super(HighlevelEventHandler, self).__init__()
        self._update_posttreatment_cursor()
        self._copied_dir_list = {}
        self._copied_dir_list.clear()
        
        self._block_releases_while_hashing = False
        self._stop_handler = threading.Event()
        self.name = 'Highlevel local event handler'


    def stop(self):
        """
        Set the :py:class:`threading.Event` so that the handler thread
        knows that it has to stop running. Call this method when you want 
        to preperly stop the handler. The handler will then stop a few seconds 
        afterwards.
        """
        self._stop_handler.set()
    
    def _update_posttreatment_cursor(self):
        """
        Private method updating the last time a post-treatment occurs for
        the watched directory. This last time cursor, combined with the 
        :py:attr:`POSTTREATMENT_TIME_LIMIT` class attribute, defines whether
        current queued events are ready to be emitted by the handler or not.
        """
        self._latest_highlevel_posttreatment = datetime.datetime.now()

    @staticmethod
    def _len_list_dir(absolute_path:str) -> int:
        """
        Private method counting the number of files and directories contained
        under the ``absolute_path`` path. Returns ``None`` if path does not exist.
        """
        try:
            return len(os.listdir(absolute_path))
        except:
            return None
    
    @staticmethod
    def _check_empty_src_dest_folder(abs_src_path:str, abs_dest_path:str) -> bool:
        """
        Private method checking that ``abs_src_path`` and ``abs_dest_path`` 
        paths contains the same files and the same directories (recursively). 
        The relative paths (including the file or dir basename) have to be 
        the same.
        """
        for root, dirs, files in os.walk(abs_src_path):  
            for i in dirs:
                abs_dest_path_i = os.path.join(abs_dest_path, i)
                if not os.path.exists(abs_dest_path_i):
                    return False
            for i in files:
                abs_dest_path_i = os.path.join(abs_dest_path, i)
                if not os.path.exists(abs_dest_path_i):
                    return False
        for root, dirs, files in os.walk(abs_dest_path):  
            for i in dirs:
                abs_src_path_i = os.path.join(abs_src_path, i)
                if not os.path.exists(abs_src_path_i):
                    return False
            for i in files:
                abs_src_path_i = os.path.join(abs_src_path, i)
                if not os.path.exists(abs_src_path_i):
                    return False
        return True
                
    
    def _posttreat_copied_folder(self):
        """
        Private helper method identifying possible matches between a list of copied events
        (from ``self.events_list``)`and the already-existing folders in the watched directory.
        If one of the already-existing folder contains a list of file and folder, each of them 
        corresponding to a related copied event, this helper removes every corresponding 
        copied events from the queue, and then creates a new copied event 
        :py:attr:`~lazydog.events.LazydogEvent.EVENT_TYPE_COPIED` related to their 
        common parent folder. As for a simple file copied event, multiple sources can be
        append to the new copied event, in order to remind all the potential source
        folder (this is sometimes useful if we need to recursively do the same post-treatment with
        the parent folder).
        """
        if len(self._copied_dir_list) == 0:
            return
        
        # Cleaning...
        for k, t in self._copied_dir_list.items():
            if datetime.datetime.now() - t > datetime.timedelta(minutes=20):
                self._copied_dir_list.pop(k)
        
        to_paths = {}
        
        # General idea is to identify if all the file from a same source folder have 
        # each of them generated copied_events to the same destination folder
         
        # For each parent_to_path, get the possible parent_src_paths
        for e in [x for x in self.events_list if x.is_copied_event() and x.parent_rp in self._copied_dir_list]:
            if not e.parent_rp in to_paths:
                to_paths[e.parent_rp] = {}
            for sp, parent_sp in e.possible_src_paths.items():
                if to_paths[e.parent_rp].get(parent_sp) is None:
                    to_paths[e.parent_rp][parent_sp] = []
                to_paths[e.parent_rp][parent_sp].append(e)
        
        # Then, we add potential empty folders that have been copied but not transformed into copied event (because no file inside)
        for e in [x for x in self.events_list if x.is_created_event() and x.parent_rp in to_paths]:
            if e.is_empty():
                for parent_sp in to_paths[e.parent_rp]:
                    absolute_source_path = self.local_states.absolute_local_path(os.path.join(parent_sp, e.basename))
                    # if directory does not exist, count_files_in returns None
                    if e.is_directory() and LazydogEvent.count_files_in(absolute_source_path) == 0:
                        if HighlevelEventHandler._check_empty_src_dest_folder(absolute_source_path, e.absolute_ref_path):
                            to_paths[e.parent_rp][parent_sp].append(e)
                    # if file does not exist, get_file_size returns None
                    elif not e.is_directory() and LazydogEvent.get_file_size(absolute_source_path) == 0:
                        to_paths[e.parent_rp][parent_sp].append(e)
        
        # Then, we check if any created folder event corresponds to the parent_to_paths
        recurse = False
        for tp in to_paths:
            dir_created_event = next(iter([x for x in self.events_list if x.is_dir_created_event() and x.ref_path == tp]), None)
            #for sp in to_paths[tp]:
            potential_sp = [x for x in to_paths[tp] 
                 if (len(to_paths[tp][x]) == HighlevelEventHandler._len_list_dir(self.local_states.absolute_local_path(x)) and 
                     len(to_paths[tp][x]) == HighlevelEventHandler._len_list_dir(self.local_states.absolute_local_path(tp)))]
            # TODO : possibility to simplify again because this loop seems not needed...
            for sp in potential_sp:
                #logging.debug("Posttreating copied folder: %s - %s - %s", len(to_paths[tp][sp]), HighlevelEventHandler._len_list_dir(sp), HighlevelEventHandler._len_list_dir(tp))
                #if (len(to_paths[tp][sp]) == HighlevelEventHandler._len_list_dir(self.local_states.absolute_local_path(sp)) and 
                #    len(to_paths[tp][sp]) == HighlevelEventHandler._len_list_dir(self.local_states.absolute_local_path(tp))):
                    # merging the copied files under a copied folder
                    # only the first iteration will remain a created_event, following one will have been transformed into copied event
                if dir_created_event is not None:
                    # once transformed into copied event, the following will not be applied
                    if dir_created_event.is_created_event():
                        # at the end we will posttreat the parent folder
                        recurse = True
                        self._copied_dir_list[os.path.dirname(tp)] = datetime.datetime.now()
                        # for now remove any related event
                        for e in to_paths[tp][sp]:
                            # if still not transformed : means it is empty file or folder
                            if e.is_created_event():
                                # for all empty subfolders and files, update e, then remove them:
                                for ee in [x for x in self.events_list if x.is_created_event() and x.comes_after(e) and x.is_empty()]:
                                    ee.update_main_event(e)
                                    self._update_local_state(ee)
                                    self.events_list.remove(ee)
                            # add source (and transforms) 
                            e.add_source_paths_and_transforms_into_copied_event(os.path.join(sp, e.basename))
                            # update main and remove
                            if e in self.events_list:
                                e.update_main_event(dir_created_event)
                                self._update_local_state(e)
                                self.events_list.remove(e)
                        # remove folder from potentially copied... since it will be effectively transformed
                        self._copied_dir_list.pop(tp)
                        self._update_local_state(dir_created_event)
                    # transform main event
                    # TODO : the following seems not well... we are using potential_sp (instead of sp) inside a loop above potential_sp....
                    dir_created_event.add_source_paths_and_transforms_into_copied_event(potential_sp)
                    self._update_posttreatment_cursor()
                    
        # if any event has been transformed to copied event:
        if recurse:
            self._posttreat_copied_folder()
        
        
    # IMPORTANT TODO : to be protected against simultaneous get_available_events modifications...              
    def posttreat_lowlevel_event(self, local_event:LazydogEvent):
        """
        Executes the main logics of the High-level Handler. These are all the aggregation 
        rules, depending on the order of arrival of the low-level event, how to identify 
        the relation between them and when to decide to aggregate them, or to transform 
        them into a high-level `Copied` or a `Moved` event.

        Please read directly the commented code for more information about these rules. Here is 
        a summary of the execution:

        * Aggregation rules:

            * Using an :py:class:`~revised_watchdog.observers.inotify.InotifyObserver`,  \
            `Deleted` events arrive backward, which means that if you delete a directory \
            with some files inside, you will get first a `Deleted` event for the inside  \
            files then another one for their parent directory. So if we find a `Deleted` \
            event for a directory, we remove every children `Deleted` events previously  \
            queued. Note that if a `Deleted` event arrives after a `Modified` event or   \
            anything else for the same file or folder, then we just remove (or adapt)    \
            the previous related events.
            * Using an :py:class:`~revised_watchdog.observers.inotify.InotifyObserver`,  \
            `Moved` events are the most simple to post-treat: if you move a folder with  \
            sub-files, you only get one low-level event. So nothing to aggregate here... \
            The only thing is when a `Moved` event is rapidly succeding a `Created` event \
            (or anything else), then you have to adapt the original event in the queue.
            * `Modified` events are easy to aggregate to other ones. They are often      \
            meaningless, since a low-level `Whatever` event often comes with one or more \
            `Modified` events, so we often just ignore these `Modified` events... Note   \
            that when you copy or create a large file, you will get multiple low-level   \
            `Modified` events per seconds that you will have to ignore (since you want   \
            to do a high-level lazy observer).

        * If the new event is not related to any other already-listed events, thent it is \
        added to the queue as a new high-level event.
        * Transformation of `Created` events into `Copied` ones, if one or more potentiel \
        sources have been found for the `Created` event. The identification of the sources \
        is based on the :py:attr:`~lazydog.events.LazydogEvent.file_size`, the             \
        :py:attr:`~lazydog.events.LazydogEvent.file_mtime` and the                         \
        :py:attr:`~lazydog.events.LazydogEvent.file_hash` attributes. The first step concerns \
        only the files. Then at the end, if any event has been transformed into a `Copied` \
        one, the :py:meth:`_posttreat_copied_folder` helper method is called.

        """
        
        # posttreat file copy is done at the end of this method
        copy_event_to_posttreat = None
        
        # created events
        if local_event.is_created_event():
            copy_event_to_posttreat = local_event
                       
        # deleted events arrive backward
        if local_event.is_deleted_event():
            for e in [x for x in reversed(self.events_list) if x.is_deleted_event()]:
                if local_event.comes_before(e):
                    e.update_main_event(local_event)
                    self.events_list.remove(e)
                    local_event.is_related = False
                    self._update_posttreatment_cursor()
                elif local_event.has_same_path_than(e):
                    local_event.update_main_event(e)
                    self._update_posttreatment_cursor()
            for e in [x for x in reversed(self.events_list) if x.has_same_path_than(local_event)]:
                if e.is_created_event() or e.is_copied_event() or e.is_modified_event():
                    e.update_main_event(local_event)
                    self.events_list.remove(e)
                    local_event.is_related = False
                    local_event.is_irrelevant = True
                    self._update_posttreatment_cursor()
                elif e.is_moved_event():
                    e.update_main_event(local_event)
                    local_event.path = e.path
                    local_event.is_related = False
                    self.events_list.remove(e)
                    self._update_posttreatment_cursor()
                    
        # moving event can also arrive just after newly created or copied or moved event
        if local_event.is_moved_event():
            self._update_local_state(local_event)
            for e in [x for x in self.events_list if (x.is_created_event() or x.is_copied_event() or x.is_moved_event())]:
                if local_event.has_same_src_path_than(e):
                    local_event.update_main_event(e)
                    e.ref_path = local_event.to_path
                elif local_event.comes_after(e) and e.is_moved_event():
                    local_event.update_main_event(e)
                
        
        # modified event
        if local_event.is_modified_event():
            # we do not need notification for modified folder
            if local_event.is_directory():
                local_event.is_related = True
            else:
                for e in [x for x in reversed(self.events_list)]:
                    # modified file event related to deleted, moved or copied event
                    if e.is_deleted_event() or e.is_moved_event() or e.is_copied_event():
                        if local_event.same_or_comes_after(e):
                            local_event.update_main_event(e)
                    # modified file event related to creted or modified event
                    elif e.is_created_event() or e.is_modified_event():
                        if local_event.has_same_path_than(e):
                            local_event.update_main_event(e)
                            # in case of lastly created event, we re-check for potential copied event
                            if e.is_created_event() and local_event.is_meta_file_modified_event():
                                copy_event_to_posttreat = e
        
        # else... if event has no relation with previous ones, we add it in the list of potential high level event
        if not local_event.is_related:
            self.events_list.append(local_event)
                
        # then posttreat file copied event
        if copy_event_to_posttreat is not None:
            if copy_event_to_posttreat.file_size is not None:
                if copy_event_to_posttreat.file_size > 0:
                    if len(self.local_states.get_files_by_sizetime_key((copy_event_to_posttreat.file_size, copy_event_to_posttreat.file_mtime))) > 0:
                        self._block_releases_while_hashing = True
                        # the following command also transforms the created event into a copied one (if any src paths found)...
                        # File Hash is computed.
                        copy_event_to_posttreat.add_source_paths_and_transforms_into_copied_event(self.local_states.get_files_by_hash_key(copy_event_to_posttreat.file_hash))
                        if copy_event_to_posttreat.is_copied_event():
                            self._copied_dir_list[os.path.dirname(copy_event_to_posttreat.to_path)] = datetime.datetime.now()
                        self._update_local_state(copy_event_to_posttreat)
                        self._update_posttreatment_cursor()
                        self._block_releases_while_hashing = False
            
        # then posttreat dir copied event
        self._posttreat_copied_folder()
        
        # posttreatment cursor temporizes the delivrances of high level event to observers
        self._update_posttreatment_cursor()
            
                        
    def _update_local_state(self, event:LazydogEvent):
        if event.is_deleted_event():
            self.local_states.delete(event.ref_path)
        elif event.is_moved_event():
            self.local_states.move(event.path, event.to_path)
        else:
            # File Hash is computed if needed
            self.save_locals(event.ref_path, [event.file_hash, event.file_size, event.file_mtime])
        
                           
    def save_locals(self, file_path, file_references):
        """
        Directly modifies the local state dictionary associated to the handler, by providing 
        new reference for a file or folder. This method should be use in combination of 
        the optional parameter ``custom_intializing_values`` when calling :py:meth:`get_instance`. 
        In the case you rapidly initialize the handler with some files values, and then 
        you see that some of these values are not good or that some files are missing, 
        you can adjust by passing new values or new files with this method. The data structure 
        is the almost the same. 

        :param file_path:
            The relative path of the file or folder you want to add.
        :type file_path:
            str
        :param file_references:
            A list of 3 values in the following order: ``file_hash``, ``file_size``, ``file_mtime``
        :type file_references:
            list
        :returns: 
            ``None``
        """
        self.local_states.save(file_path, file_references[0], file_references[1], file_references[2])
        
    # IMPORTANT TODO : to be protected against simultaneous posttreat_lowlevel_event modifications... 
    def get_available_events(self) -> list:
        """
        Returns a list of high-level post-treated and ready events. Ready in the sense
        that the :py:attr:`POSTTREATMENT_TIME_LIMIT` has been reached without any new
        low-level events coming... 
        
        .. note: If you instanciated a new Handler using the :py:meth:`get_instance`
            method, then you got an already-running handler. No need to call the 
            :py:meth:`start` method again.

        """
        ready_events = []
        if not self._block_releases_while_hashing and LazydogEvent.datetime_difference_from_now(self._latest_highlevel_posttreatment) > HighlevelEventHandler.POSTTREATMENT_TIME_LIMIT:

            if all(x.idle_time() > HighlevelEventHandler.POSTTREATMENT_TIME_LIMIT for x in self.events_list):
                
                for e in self.events_list.copy():
                    if e.is_file_created_event() and e.is_empty() and e.idle_time() <= HighlevelEventHandler.CREATE_EVENT_TIME_LIMIT_FOR_EMPTY_FILES:
                        continue
                    
                    if e.is_irrelevant:
                        self.events_list.remove(e)
                        continue

                    self.events_list.remove(e)
                    ready_events.append(e)
                    
                # clean erratic modified events (when receiving a `Modified` event
                # but there is no real modification...). We can not check this kind 
                # of problem in the posttreat_lowlevel_event method, because it needs 
                # to compute the file hash (and we could receive alot of `Modified` 
                # event in the first place, that will be merged, so treating this 
                # here allow to compute the hash only once).
                for e in ready_events.copy():
                    if e.is_modified_event():
                        # File Hash is computed.
                        if ( e.file_hash == self.local_states.get_hash(e.path, compute_if_none=False) and
                            (e.file_size, e.file_mtime) == self.local_states.get_sizetime(e.path, compute_if_none=False)):
                            ready_events.remove(e)
                        else:
                            self._update_local_state(e)
                    elif e.is_created_event() or e.is_deleted_event():
                        self._update_local_state(e)
                            
        return ready_events
    
    def run(self):
        """
        Threading module method, that is executed when calling :py:meth:`start` method.
        The thread is running in a loop until you call the :py:meth:`stop` method. Until 
        then, it just check regularly if there is any new queued events emitted by the watchdog 
        oberver. If any, it post-treats it calling the  
        :py:meth:`posttreat_lowlevel_event` method.
        
        .. note: If you instanciated a new Handler using the :py:meth:`get_instance`
            method, then you got an already-running handler. No need to call the 
            :py:meth:`start` method again.

        """
        while not self._stop_handler.is_set():
            
            time.sleep(0.2) # to give some time for hashing from other threads...
            
            while not self.lowlevel_event_queue.is_empty():                

                # Post-treatment of the next lowlevel event
                lowlevel_event = self.lowlevel_event_queue.next()
                # print('+++ INCOMING LOW LEVEL EVENT +++')
                # for e in self.events_list:
                #     print('+ ' + str(e))
                # print(lowlevel_event)
                
                #logging.debug('+++' + str(lowlevel_event))
                self.posttreat_lowlevel_event(lowlevel_event)
                



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: lazydog.events
:synopsis: Definitions of the high-level lazydog events, based on the \
low-level watchdog ones, which are now aggregable and also convertible \
to copy or move events.
:author: Clément Warneys <clement.warneys@gmail.com>

Possible type of lazydog events:

* :py:attr:`~LazydogEvent.EVENT_TYPE_CREATED` for the creation of a file or folder
* :py:attr:`~LazydogEvent.EVENT_TYPE_MODIFIED` for the creation of a file or folder \
(whatever the modification concerns: metadata or content)
* :py:attr:`~LazydogEvent.EVENT_TYPE_MOVED` for the move of a file or folder
* :py:attr:`~LazydogEvent.EVENT_TYPE_COPIED` for the copy of a file or folder
* :py:attr:`~LazydogEvent.EVENT_TYPE_DELETED` for the deletion of a file or folder

.. note:: Some kind of events such as Moved and Copied have 2 path attributes:
    :py:attr:`~LazydogEvent.path` for the origin path, and 
    :py:attr:`~LazydogEvent.to_path` for the destination path. Other kinds have only 
    the :py:attr:`~LazydogEvent.path` attribute.

    The :py:attr:`~LazydogEvent.ref_path` attribute always refers 
    to the current location of the file (:py:attr:`~LazydogEvent.to_path` if any, 
    else :py:attr:`~LazydogEvent.path`). All the paths are always relative to the 
    main watched directory.

Lazydog has the ability to aggregate related low-level events. For example, in 
the case of multiple deletion events, each of one under the same parent directory, 
the lazydog handler will emit only one deletion event, with the path of the common 
parent directory.

Lazydog is also able to correlate almost simultaneous deletion and creation events
into a unique moved event, if the low-level events are related. Or mutiple creation events
into a unique copied event, if the new files and folders were already existing 
elsewhere in the main watched folder.

All these correlations are mainly done by the :py:class:`~lazydog.handlers.HighlevelEventHandler`
class, but some helper methods are defined in the :py:class:`~lazydog.events.LazydogEvent`
class such as :py:meth:`~lazydog.events.LazydogEvent.add_source_paths_and_transforms_into_copied_event`
or :py:meth:`~lazydog.events.LazydogEvent.update_main_event`.

"""

import datetime
import os

from lazydog.states import LocalState

from lazydog.revised_watchdog.events import (
    FileSystemEvent,
    EVENT_TYPE_C_MODIFIED,
    EVENT_TYPE_M_MODIFIED 
    )

from watchdog.events import (
    EVENT_TYPE_MOVED, 
    EVENT_TYPE_CREATED, 
    EVENT_TYPE_DELETED
    )

class LazydogEvent():
    """
    Main class of :py:mod:`lazydog.events` module. Initialization with a 
    low-level watchdog event that is then converted into 
    high-level lazydog event.

    .. note:: The local path of the event is referenced as a relative path
        starting from the absolute path of the watched directory. For this 
        mechanism, the Lazydog event needs a reference, which is given 
        at the initialisation with a :class:`LocalState` reference.

    :param event:
        A low-level watchdog event.
    :type event:
        :py:class:`~revised_watchdog.events.FileSystemEvent`
    :param local_states:
        The reference state of the local files in the watched directory.
        Including the absolute path of the watched directory, thus allowing to 
        manage high-level event with relative path.
    :type local_states:
         :py:class:`~lazydog.states.LocalState`
    :returns: 
        A high-level lazydog event (converted from low-level watchdog event).
    :rtype: 
        :py:class:`~lazydog.events.LazydogEvent`

    """


    EVENT_TYPE_CREATED = EVENT_TYPE_CREATED # 'created'
    """Created event type, imported from :py:mod:`watchdog` module"""

    EVENT_TYPE_DELETED = EVENT_TYPE_DELETED # 'deleted'
    """Deleted event type, imported from :py:mod:`watchdog` module"""

    EVENT_TYPE_MOVED = EVENT_TYPE_MOVED # 'moved'
    """Moved event type, imported from :py:mod:`watchdog` module"""

    EVENT_TYPE_C_MODIFIED = EVENT_TYPE_C_MODIFIED # 'modified' # content modification
    """Content modified event type, imported from :py:mod:`lazydog.revised_watchdog` module"""

    EVENT_TYPE_M_MODIFIED = EVENT_TYPE_M_MODIFIED # 'metadata' # metadata modification
    """Metadata modified event type, imported from :py:mod:`lazydog.revised_watchdog` module"""
        
    EVENT_TYPE_COPIED = 'copied'
    """
    New kind of event, that does not exist in watchdog python module.
    Copied event can only be obtained by transforming Created events. 
    The transformation decision is made by the 
    :py:class:`~lazydog.handlers.HighlevelEventHandler` and is based on 
    the existing files or folders in the watched directory.
    """
    
    def __init__(self, event:FileSystemEvent, local_states:LocalState):
        # Dating now
        self.event_date = datetime.datetime.now()
        
        # Saving LocalState Reference
        self.local_states = local_states
        
        # FileSystemEvent definitions
        self.type = event.event_type
        self.is_dir = event.is_directory
        
        # Path and local file handling
        self.path = self.local_states.relative_local_path(event.src_path)
        self.to_path = self.local_states.relative_local_path(event.dest_path) if self.has_dest() else None
        self._reset_ref_paths()
            
        # Local file or folder computations
        self._reset_file_infos()
        
        # Helpers for high-level event identification
        self.possible_src_paths = {}
        self.possible_src_paths.clear()
        self.related_events = []
        self.related_events.clear()
        self.related_events.append(self)
        self.first_event_date = self.event_date
        self.latest_event_date = self.event_date
        self.latest_reworked_date = datetime.datetime.now()
        self.is_related = False
        self.is_irrelevant = False

        
    def __str__(self):
        if self.is_moved_event() or self.is_copied_event():
            return ('logged[' + str(self.event_date) + '] ' +
                    self.type + ": '" + self.path + "' " +
                    "to '" + self.to_path + "' " + 
                    'mtime[' + str(self.file_mtime) + '] ' +
                    'size[' + str(self.file_size) + '] ' +
                    'inode[' + str(self.file_inode) + '] ' +
                    ('irrelevant ' if self.is_irrelevant else ''))
        else:
            return ('logged[' + str(self.event_date) + '] ' +
                    self.type + ": '" + self.path +  "' " + 
                    'mtime[' + str(self.file_mtime) + '] ' +
                    'size[' + str(self.file_size) + '] ' +
                    'inode[' + str(self.file_inode) + '] ' +
                    ('irrelevant ' if self.is_irrelevant else ''))
            
    
    # managing weird phenomenom when approahing the root of the watched dir.
    def _correct_path_value(self, value:str) -> str:
        return '/' if value == '/.' else value
    
    @property
    def path(self) -> str:
        """Origin path of the event."""
        return self._path

    @path.setter
    def path(self, value:str):
        self._path = self._correct_path_value(value)
        self._reset_ref_paths()
        
    @property
    def to_path(self) -> str:
        """Destination path of the event, if any, else ``None``."""
        return self._to_path

    @to_path.setter
    def to_path(self, value:str):
        self._to_path = self._correct_path_value(value)
        self._reset_ref_paths()
    
    
    def _reset_ref_paths(self):
        self._ref_path = None
        self._absolute_ref_path = None
    
    @property
    def ref_path(self) -> str:
        """
        Refers to the current location of the file or the event,
        which is :py:attr:`to_path` if any, else :py:attr:`path`.
        """
        if self._ref_path is None:
            self._ref_path = self.to_path if self.has_dest() else self.path
        return self._ref_path
    
    @ref_path.setter
    def ref_path(self, value:str):
        if self.has_dest():
            self.to_path = value
        else:
            self.path = value
    
    # Parent directory of destination directory
    @property
    def parent_rp(self) -> str:
        """
        Refers to the directory name of the event.
        If the directory name is already the main watched 
        directory, ``None`` is returned.
        """
        if self.ref_path == '/':
            return None
        else:
            return os.path.dirname(self.ref_path)
        
    # Filename or dirname
    @property
    def basename(self) -> str:
        """Returns the filename or directory name of the related file or dir."""
        return os.path.basename(self.ref_path)
    
    @property
    def absolute_ref_path(self) -> str:
        """Returns the absolute path of the current location of the file or dir."""
        if self._absolute_ref_path is None:
            self._absolute_ref_path = self.local_states.absolute_local_path(self.ref_path)
        return self._absolute_ref_path
    
    
    def is_directory(self) -> bool:
        """Returns ``True`` if the event is related to a directory."""
        return self.is_dir
    
    # Check if same type (only type, without considering Dir)
    def is_same_type_than(self, event) -> bool:
        return self.type == event.type
    
    def is_moved_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir move."""
        return self.type == LazydogEvent.EVENT_TYPE_MOVED
    
    def is_dir_moved_event(self) -> bool:
        """Returns ``True`` if the event is a dir move."""
        return self.is_moved_event() and self.is_directory()
    
    def is_deleted_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir deletion."""
        return self.type == LazydogEvent.EVENT_TYPE_DELETED
    
    def is_dir_deleted_event(self) -> bool:
        """Returns ``True`` if the event is a dir deletion."""
        return self.is_deleted_event() and self.is_directory()
    
    def is_created_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir creation."""
        return self.type == LazydogEvent.EVENT_TYPE_CREATED
    
    def is_dir_created_event(self) -> bool:
        """Returns ``True`` if the event is a dir creation."""
        return self.is_created_event() and self.is_directory()
    
    def is_file_created_event(self) -> bool:
        """Returns ``True`` if the event is a file creation."""
        return self.is_created_event() and not self.is_directory()
    
    def is_copied_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir copy."""
        return self.type == LazydogEvent.EVENT_TYPE_COPIED
    
    def is_modified_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir modification."""
        return self.is_meta_modified_event() or self.is_data_modified_event()
    
    def is_meta_modified_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir modification of the metadata only."""
        return self.type == LazydogEvent.EVENT_TYPE_M_MODIFIED
    
    def is_data_modified_event(self) -> bool:
        """Returns ``True`` if the event is a file or dir modification of the content."""
        return self.type == LazydogEvent.EVENT_TYPE_C_MODIFIED 
    
    def is_file_modified_event(self) -> bool:
        """Returns ``True`` if the event is a file modification."""
        return self.is_modified_event() and not self.is_directory()
    
    def is_meta_file_modified_event(self) -> bool:
        """Returns ``True`` if the event is a file modification of the metadata only."""
        return self.is_meta_modified_event() and not self.is_directory()
    
    def is_data_file_modified_event(self) -> bool:
        """Returns ``True`` if the event is a file modification of the content."""
        return self.is_data_modified_event() and not self.is_directory()
    
    def is_dir_modified_event(self) -> bool:
        """Returns ``True`` if the event is a dir modification."""
        return self.is_modified_event() and self.is_directory()
    
    def has_dest(self) -> bool:
        """
        Returns ``True`` if the event has a destination path 
        (i.e. if it's a Moved or Copied event).
        """
        return self.is_moved_event() or self.is_copied_event()
    
    def has_same_mtime_than(self, previous_event) -> bool:
        """Returns ``True`` if the event has the same modification time than the event in parameter."""
        return self.file_mtime == previous_event.file_mtime
    
    # Check size 
    def has_same_size_than(self, event) -> bool:
        """Returns ``True`` if the event has the same size than the event in parameter."""
        return self.file_size == event.file_size
    
    # Check paths
    def has_same_path_than(self, event) -> bool:
        """
        Returns ``True`` if the event has the same :py:attr:`ref_path` 
        than the event in parameter.
        
        If both events have destination path, source paths
        are compared too.
        """
        if self.has_dest() and event.has_dest():
            return self.path == event.path and self.to_path == event.to_path
        else:
            return self.ref_path == event.ref_path
    
    def has_same_src_path_than(self, event) -> bool:
        """
        Returns ``True`` if the :py:attr:`path` of event is the same 
        than the :py:attr:`ref_path` of the event in parameter.
        """
        return self.path == event.ref_path
        
    
    
    @staticmethod
    def p1_comes_after_p2(p1:str, p2:str) -> bool:
        """
        p1 and p2 are both paths (str format). This method is
        a basic comparison method to check if the first parameter p1
        is striclty a parent path of the second parameter p2. 

        Returns ``False`` if both paths are identical.
        """
        if not p1.endswith('/'):
            p1 = p1 + '/'
        if not p2.endswith('/'):
            p2 = p2 + '/'
        return p1.startswith(p2) and p1 != p2
    
    @staticmethod
    def p1_comes_before_p2(p1:str, p2:str) -> bool:
        """Same than :py:meth:`p1_comes_after_p2` method, but opposite result.""" 
        return LazydogEvent.p1_comes_after_p2(p2, p1)
    
    def comes_before(self, event) -> bool:
        """Same than :py:meth:`comes_after` method, but opposite result.""" 
        return event.comes_after(self)
            
    def same_or_comes_before(self, event) -> bool:
        """Same than :py:meth:`comes_before` method, but also ``True`` when both events have identical paths.""" 
        return self.comes_before(event) or self.has_same_path_than(event)
    
    def comes_after(self, event, complete_check:bool=True) -> bool:
        """
        Same result than :py:meth:`p1_comes_after_p2`,
        comparing current event :py:attr:`ref_path` path (as p1), to the 
        :py:attr:`ref_path` path of the event in parameter (as p2).

        If both events have a destination path, source paths
        are compared too.

        Returns ``False`` if both paths are identical.
        """ 
        if complete_check and self.has_dest() and event.has_dest():
            return LazydogEvent.p1_comes_after_p2(self.path, event.path) and LazydogEvent.p1_comes_after_p2(self.to_path, event.to_path)
        else:
            return LazydogEvent.p1_comes_after_p2(self.ref_path, event.ref_path)
        
    def same_or_comes_after(self, event) -> bool:
        """Same than :py:meth:`comes_after` method, but also ``True`` when both events have identical paths.""" 
        return self.comes_after(event) or self.has_same_path_than(event)
                
    
    @staticmethod
    def datetime_difference_from_now(dt:datetime.datetime) -> datetime.datetime:
        """ 
        Returns :py:class:`datetime.datetime` object representing time difference 
        between the datetime in parameter, and now.
        """
        return datetime.datetime.now() - dt
    
    def idle_time(self) -> datetime.datetime:
        """ 
        Returns time difference between last time this event has been updated
        and now. 

        .. note:: Event updates occur when the event is aggregated to another 
            related event, or also when the event is transformed into a copied
            or a moved one...
        """
        return LazydogEvent.datetime_difference_from_now(self.latest_reworked_date)
                                                      
    @property
    def file_hash(self) -> str:
        """
        Returns the file hash of the file related to the event if any, else ``None``.
        File hash value is saved into a private variable, in order to avoid useless
        computation time...
        """
        if self._file_hash is None:
            if self.is_directory():
                self._file_hash = self.local_states.DEFAULT_DIRECTORY_VALUE
            else:
                try:
                    # Here we don't use local_states which is considered as historic value of hash. 
                    # We need to compute new hash value (to be able to compare it to historic one)
                    self._file_hash = self.local_states.hash_function(self.absolute_ref_path)
                except:
                    self._file_hash = None
        return self._file_hash
    
              
    @staticmethod
    def count_files_in(absolute_dir_path:str) -> int:
        """
        Counts all non-empty (file size > 0) files in ``absolute_dir_path`` 
        directory and all its sub-directories. Returns ``None`` if 
        the ``absolute_dir_path`` is not a directory.

        .. note:: Be careful: ``absolute_dir_path`` has to represent 
            absolute path (not a relative one).
        """
        qty = None
        try:
            if os.path.isdir(absolute_dir_path):
                qty = 0  
                for root, dirs, files in os.walk(absolute_dir_path):
                    for f in files:
                        if os.path.getsize(os.path.join(root, f)) > 0:
                            qty = qty + 1
        except:
            pass
        return qty  
    
    @property
    def dir_files_qty(self) -> int:
        """
        Counts all non-empty (file size > 0) files in the related path
        of the event, and all its sub-directories. Returns ``None`` 
        if the event is not related to a directory.
        """
        if self._dir_files_qty is None and self.is_directory():
            self._dir_files_qty = LazydogEvent.count_files_in(self.absolute_ref_path)
        return self._dir_files_qty
    
    @staticmethod
    def get_file_size(absolute_file_path:str) -> int:
        """Returns the size of the file at the specified absolute path if any, else ``None``."""
        try:
            return os.path.getsize(absolute_file_path) if not os.path.isdir(absolute_file_path) else None
        except:
            return None
        
    @property
    def file_size(self) -> int:
        """
        Size of the file related to the event if any, else ``None``.
        File size value is saved in a private variable, in order to avoid useless
        sollicitation of file-system.
        """
        if self._file_size is None and not self.is_directory():
            self._file_size = LazydogEvent.get_file_size(self.absolute_ref_path)
        return self._file_size
    
    def is_empty(self) -> bool:
        """
        Returns ``True`` if the event is related to an empty directory, 
        or if the event is related to an empty file (size = 0).
        """
        if self.is_directory():
            return self.dir_files_qty == 0
        else:
            return self.file_size == 0
    
    @property
    def file_mtime(self) -> float:
        """
        Last modification time of the file related to the event if any, else ``None``.
        File modification time value is saved in a private variable, in order to avoid useless
        sollicitation of file-system.
        """
        if self._file_mtime is None:
            try:
                self._file_mtime = round(os.path.getmtime(self.absolute_ref_path), 3)
            except:
                self._file_mtime = None
        return self._file_mtime
    
    @property
    def file_inode(self) -> int:
        """
        Inode of the file related to the event if any, else ``None``.
        Inode value is saved in a private variable, in order to avoid useless
        sollicitation of file-system.

        .. note:: This property seems now useless, and could be deprecated.
        """
        if self._file_inode is None:
            try:
                self._file_inode = os.stat(self.absolute_ref_path).st_ino
            except:
                self._file_inode = None
        return self._file_inode
        
    
    def _reset_file_infos(self):
        # reset
        self._file_inode = None
        self._file_size = None
        self._file_mtime = None
        self._file_hash = None
        self._dir_files_qty = None
        # compute again
        self.file_mtime
        self.file_size
        self.file_inode
        self.dir_files_qty
        
    
    def update_main_event(self, main_event):    
        """
        High level helper method to facilitate the work of the 
        :py:class:`~lazydog.handlers.HighlevelEventHandler`. 
        When different events are identified as related ones, this method 
        is merging the current event in the main one (in parameter). 

        General idea is to update paramters of the main event, such as 
        :py:attr:`file_inode`, :py:attr:`file_mtime`, :py:attr:`file_size`, 
        :py:attr:`file_hash`, and also the dates of occurence (which are needed
        to manage an aggregation time limit).

        Each related events, including the main event itself, are all listed 
        in :py:attr:`related_events` list, to keep track of them.
        """

        # usual aggregations
        for e in self.related_events:
            main_event.related_events.append(e)
        # re-init
        if self.is_modified_event() and not self.is_directory() and not main_event.is_directory():
            if main_event._file_mtime != self._file_mtime or main_event._file_size != self._file_size:
                main_event._file_inode = self._file_inode
                main_event._file_mtime = self._file_mtime
                main_event._file_size = self._file_size
                main_event._file_hash = self._file_hash
        if main_event.is_directory():
            self._dir_files_qty = None
        # timing    
        if main_event.latest_event_date < self.latest_event_date:
            main_event.latest_event_date = self.latest_event_date
        if main_event.first_event_date > self.first_event_date:
            main_event.first_event_date = self.first_event_date
        main_event.latest_reworked_date = datetime.datetime.now()
        # is now related
        self.is_related = True
        main_event.is_related = True
    
    # looking for the most probable source aamong a list of potential sources, based on the basename of the destination file or folder
    @staticmethod
    def _get_most_potential_source(src_paths:set, dest_path:str) -> str:
        """
        High level private helper method that identifies the best match 
        when there are multiple possibilities to aggregate creations events 
        in a copied one. 

        To be more accurate, when there is a new file creation, it can correspond
        to an existing file by matching its size, its modification date, and eventually 
        its hash. Actually, there could be multiple matches (if you copy the same file 
        multiple times). The idea is to get the more likely source file, based on the
        filename of the different sources, and comparing to the desination filename.

        Usually, if the source filename is something like "foo.txt", the destination 
        filename will be something like "Copy of foo.txt" or "foo-copy.txt" or anything
        approaching, depending on the operating system that did the copy. The goal
        is to catch the "foo.txt" source filename preferentially before any other one.
        """
        most_potential_sources = [x for x in src_paths if os.path.splitext(os.path.basename(x))[0] in os.path.splitext(os.path.basename(dest_path))[0]]
        most_potential_sources.sort(key = lambda x: -len(x))
        return most_potential_sources[0] if most_potential_sources else next(iter(src_paths))

    def add_source_paths_and_transforms_into_copied_event(self, src_paths:set):
        """
        High level helper method to facilitate the work of the 
        :py:class:`~lazydog.handlers.HighlevelEventHandler`. 
        When a creation event is actually identified as a copied one, 
        this method is transforming the current event in a copied one. 
 
        The old :py:attr:`path` attribute is converted into a :py:attr:`to_path`
        one. And the :py:attr:`path` id filled with one of the identified possible 
        source paths (this identification is the job of the 
        :py:class:`~lazydog.handlers.HighlevelEventHandler`).

        To get prepared to potential future aggregation of multiple copied 
        events (for example in the case of a copied directory), we need to keep 
        track of all the possible source paths which are then saved into 
        a :py:attr:`possible_src_paths` attribute.
        """
        if src_paths.__class__.__name__ == str.__name__:
            src_paths = set([src_paths])
        for sp in src_paths:
            if not self.is_copied_event():
                self.type = LazydogEvent.EVENT_TYPE_COPIED
                self.to_path = self.path
                self.path = LazydogEvent._get_most_potential_source(src_paths, self.to_path)
            # save all potential parent source path, for future use
            if os.path.basename(sp) == os.path.basename(self.to_path):
                self.possible_src_paths[sp] = os.path.dirname(sp) if sp != '/' else None
            


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
# Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: revised_watchdog.observers.inotify
:synopsis: ``inotify(7)`` based emitter implementation, enhanced implementation of original watchdog one.
:author: Sebastien Martini <seb@dbzteam.org>
:author: Luke McCarthy <luke@iogopro.co.uk>
:author: yesudeep@google.com (Yesudeep Mangalapilly)
:author: Tim Cuthbertson <tim+github@gfxmonk.net>
:author: Clément Warneys <clement.warneys@gmail.com>
:platforms: Linux 2.6.13+.

This module is overloading the original :py:mod:`watchdog.observers.inotify` module 
by revising and completing it. Please read original **watchdog** project 
documentation for more information: https://github.com/gorakhargosh/watchdog

The main changes concern some methods in the :py:class:`InotifyEmitter` class:

* :py:meth:`~InotifyEmitter.on_thread_start` This method now uses revised \
:py:class:`~lazydog.revised_watchdog.observers.inotify_buffer.InotifyBuffer`.
* :py:meth:`~InotifyEmitter.queue_events` This method has been simplified in order to reduce \
the number of emitted low-level events, in comparison with \
original watchdog module. 


"""

from watchdog.observers.inotify import (
    InotifyEmitter, 
    InotifyObserver
    )

from watchdog.observers.api import (
    BaseObserver,
    DEFAULT_OBSERVER_TIMEOUT
)

from watchdog.utils import unicode_paths

from watchdog.events import (
    FileMovedEvent,
    DirMovedEvent,
    FileCreatedEvent,
    DirCreatedEvent,
    FileDeletedEvent,
    DirDeletedEvent,
    )


from lazydog.revised_watchdog.observers.inotify_buffer import InotifyBuffer

from lazydog.revised_watchdog.events import (
    TrueFileModifiedEvent,
    MetaFileModifiedEvent,
    TrueDirModifiedEvent,
    MetaDirModifiedEvent
    )

class InotifyEmitter(InotifyEmitter):
    """
    inotify(7)-based event emitter. Revised package mainly 
    concerns :py:meth:`queue_events` method, thus covering 
    specific needs of lazydog package.

    :param event_queue:
        The event queue to fill with events.
    :type event_queue:
        :py:class:`watchdog.events.EventQueue`
    :param watch:
        A watch object representing the directory to monitor.
    :type watch:
        :py:class:`watchdog.observers.api.ObservedWatch`
    :param timeout:
        Read events blocking timeout (in seconds).
    :type timeout:
        float
    """

    def on_thread_start(self):
        path = unicode_paths.encode(self.watch.path)
        self._inotify = InotifyBuffer(path, self.watch.is_recursive)


    def queue_events(self, timeout, full_events=False):
        """
        This method is classifying the events received from Inotify into
        watchdog events type (defined in :py:mod:`watchdog.events` module).

        :param timeout:
            Unused param (from watchdog original package).
        :type timeout:
            float
        :param full_events:
            If ``True``, then the method will report unmatched move 
            events as separate events. This means that if 
            ``True``, a file move event from outside the watched directory
            will result in a :py:class:`watchdog.events.FileMovedEvent` event, with no origin. Else 
            (if ``False``), it will result in a :py:class:`watchdog.events.FileCreatedEvent` event.
            This behavior is by default only called by a :py:class:`InotifyFullEmitter`.
        :type full_events:
            boolean
        """
        with self._lock:

            event = self._inotify.read_event()
            if event is None:
                return
            if isinstance(event, tuple):
                move_from, move_to = event
                src_path = self._decode_path(move_from.src_path)
                dest_path = self._decode_path(move_to.src_path)
                cls = DirMovedEvent if move_from.is_directory else FileMovedEvent
                self.queue_event(cls(src_path, dest_path))
                #===============================================================
                # self.queue_event(MetaDirModifiedEvent(os.path.dirname(src_path)))
                # self.queue_event(MetaDirModifiedEvent(os.path.dirname(dest_path)))
                #===============================================================
                #===============================================================
                # if move_from.is_directory and self.watch.is_recursive:
                #     for sub_event in generate_sub_moved_events(src_path, dest_path):
                #         self.queue_event(sub_event)
                #===============================================================
                return

            src_path = self._decode_path(event.src_path)
            if event.is_moved_to:
                if (full_events):
                    cls = DirMovedEvent if event.is_directory else FileMovedEvent
                    self.queue_event(cls(None, src_path))
                else:
                    cls = DirCreatedEvent if event.is_directory else FileCreatedEvent
                    self.queue_event(cls(src_path))
                #===============================================================
                # self.queue_event(DirModifiedEvent(os.path.dirname(src_path)))
                #===============================================================
                #===============================================================
                # if event.is_directory and self.watch.is_recursive:
                #     for sub_event in generate_sub_created_events(src_path):
                #         self.queue_event(sub_event)
                #===============================================================
            elif event.is_attrib:
                cls = MetaDirModifiedEvent if event.is_directory else MetaFileModifiedEvent
                self.queue_event(cls(src_path))
            elif event.is_modify:
                cls = TrueDirModifiedEvent if event.is_directory else TrueFileModifiedEvent
                self.queue_event(cls(src_path))
            elif event.is_delete or (event.is_moved_from and not full_events):
                cls = DirDeletedEvent if event.is_directory else FileDeletedEvent
                self.queue_event(cls(src_path))
                #===============================================================
                # self.queue_event(DirModifiedEvent(os.path.dirname(src_path)))
                #===============================================================
            elif event.is_moved_from and full_events:
                cls = DirMovedEvent if event.is_directory else FileMovedEvent
                self.queue_event(cls(src_path, None))
                #===============================================================
                # self.queue_event(DirModifiedEvent(os.path.dirname(src_path)))
                #===============================================================
            elif event.is_create:
                cls = DirCreatedEvent if event.is_directory else FileCreatedEvent
                self.queue_event(cls(src_path))
                #===============================================================
                # self.queue_event(DirModifiedEvent(os.path.dirname(src_path)))
                #===============================================================


class InotifyObserver(InotifyObserver):
    """
    Observer thread that schedules watching directories and dispatches
    calls to event handlers. 

    Please note that his class remains unmodified 
    in revised_watchdog package. Only the :py:meth:`__init__` method is overided 
    in order it uses the new definition of :py:class:`InotifyEmitter` class.
    """

    def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT, generate_full_events=False):
        if (generate_full_events):
            BaseObserver.__init__(self, emitter_class=InotifyFullEmitter, timeout=timeout)
        else:
            BaseObserver.__init__(self, emitter_class=InotifyEmitter, timeout=timeout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
# Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: revised_watchdog.observers.inotify_c
:author: yesudeep@google.com (Yesudeep Mangalapilly)
:author: Clément Warneys <clement.warneys@gmail.com>

This module is overloading the original :py:mod:`watchdog.observers.inotify_c` module 
by revising and completing it. Please read original **watchdog** project 
documentation for more information: https://github.com/gorakhargosh/watchdog

Fundamental changes and corrections have been brought to the original :py:class:`Inotify` 
class, whose behaviour was not correct when moving or deleting sub-directories.

"""

import os
import errno

from watchdog.observers.inotify_c import (
    Inotify, 
    InotifyEvent, 
    InotifyConstants,
    DEFAULT_EVENT_BUFFER_SIZE
    )


class Inotify(Inotify):
    """
    Linux inotify(7) API wrapper class. 

    With modified :py:meth:`read_events` method, 
    and added :py:meth:`_remove_watch_bookkeeping` method, 
    thus covering specifics needs of lazydog.

    :param path:
        The directory path for which we want an inotify object.
    :type path:
        bytes
    :param recursive:
        ``True`` if subdirectories should be monitored. ``False`` otherwise.
    :type recursive:
        boolean
    """
    
    def _remove_watch_bookkeeping(self, path):
        wd = self._wd_for_path.pop(path)
        del self._path_for_wd[wd]
        return wd


    def read_events(self, event_buffer_size=DEFAULT_EVENT_BUFFER_SIZE):
        """
        Reads events from inotify and yields them to the Inotify buffer.
        This method has been largely modified from original watchdog module...
        Thus preventing from unwanted behaviour.
        """
        # HACK: We need to traverse the directory path
        # recursively and simulate events for newly
        # created subdirectories/files. This will handle
        # mkdir -p foobar/blah/bar; touch foobar/afile

        def _recursive_simulate(src_path, event_type, cookie=0):
            events = []
            for root, dirnames, filenames in os.walk(src_path):
                for dirname in dirnames:
                    try:
                        full_path = os.path.join(root, dirname)
                        wd_dir = self._add_watch(full_path, self._event_mask)
                        e = InotifyEvent(
                            wd_dir, event_type | InotifyConstants.IN_ISDIR, cookie, dirname, full_path)
                        events.append(e)
                    except OSError:
                        pass
                for filename in filenames:
                    full_path = os.path.join(root, filename)
                    wd_parent_dir = self._wd_for_path[os.path.dirname(full_path)]
                    e = InotifyEvent(
                        wd_parent_dir, event_type, cookie, filename, full_path)
                    events.append(e)
            return events

        event_buffer = None
        while True:
            try:
                event_buffer = os.read(self._inotify_fd, event_buffer_size)
            except OSError as e:
                if e.errno == errno.EINTR:
                    continue
            break

        with self._lock:

            event_list = []
            for wd, mask, cookie, name in Inotify._parse_event_buffer(event_buffer):

                if wd == -1:
                    continue
                wd_path = self._path_for_wd[wd]
                src_path = os.path.join(wd_path, name) if name else wd_path #avoid trailing slash
                inotify_event = InotifyEvent(wd, mask, cookie, name, src_path)
                move_src_path = None
                #print(inotify_event)
                
                if inotify_event.is_moved_from:
                    self.remember_move_from_event(inotify_event)
                elif inotify_event.is_moved_to:
                    move_src_path = self.source_for_move(inotify_event)
                    if move_src_path is not None:
                        # Adjusting existing watcher paths
                        for moved_path in [x for x in self._wd_for_path if x.startswith(move_src_path)]:
                            new_path = moved_path.replace(move_src_path, inotify_event.src_path, 1)
                            moved_wd = self._wd_for_path[moved_path]
                            self._wd_for_path.pop(moved_path)
                            self._wd_for_path[new_path] = moved_wd
                            self._path_for_wd[moved_wd] = new_path
                        
                    #===========================================================
                    # src_path = os.path.join(wd_path, name)
                    # inotify_event = InotifyEvent(wd, mask, cookie, name, src_path)
                    #===========================================================

                if inotify_event.is_ignored:
                    # Clean up book-keeping for deleted watches.
                    self._remove_watch_bookkeeping(src_path)
                    continue

                event_list.append(inotify_event)

                if (self.is_recursive and inotify_event.is_directory): 
                    
                    if inotify_event.is_create:
                        
                        # Putting newly created dir under observation
                        try:
                            self._add_watch(src_path, self._event_mask)
                        except OSError:
                            continue
                        
                        # The following generate elements that would have been created before 
                        # watcher had time to register...
                        event_list.extend(_recursive_simulate(src_path, InotifyConstants.IN_CREATE))
                    
                    
                    elif inotify_event.is_moved_to and move_src_path is None:
                    
                        # When a directory from another part of the
                        # filesystem is moved into a watched directory, this
                        # will not generate events for the directory tree.
                        # We need to coalesce IN_MOVED_TO events and those
                        # IN_MOVED_TO events which don't pair up with
                        # IN_MOVED_FROM events should be marked IN_CREATE
                        # instead relative to this directory.
                        
                        # Putting newly created dir under observation
                        try:
                            self._add_watch(src_path, self._event_mask)
                        except OSError:
                            continue
                        
                        # The following generate elements that would have been created before 
                        # watcher had time to register...
                        event_list.extend(_recursive_simulate(src_path, InotifyConstants.IN_MOVED_TO, -1))
                        
                        

        return event_list




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
# Copyright 2017 Dropbox, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: lazydog.dropbox_content_hasher
:synopsis: Function to get hash of a file, based on dropbox api hasher.
:author: Dropbox, Inc.
:author: Clément Warneys <clement.warneys@gmail.com>

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib
# import six

import time
import os
import logging


def default_hash_function(absolute_path:str, default_directory_hash:str='DIR'):
    """
    Main function in this module that returns the 
    dropbox-like hash of any local file. If the local path does not exist, 
    ``None`` is returned. If the local path is a directory, the 
    ``default_directory_hash`` parameter is returned, or the default 
    string "DIR". 

    :param absolute_path:
        The absolute local path of the file or directory.
    :type absolute_path:
        str
    :param default_directory_hash:
        *Optional*. The returned value in case the absolute path is a directory.
    :type absolute_path:
        str
    :returns: 
        The hash of the file or directory located in ``absolute_path``. 
        The hash is computed based on the default Dropbox API hasher. 
        ``None`` if absolute local path does not exist.
    :rtype: 
        str

    """

    _hash = None
    try:
        duration = time.perf_counter()
        # Directory
        if os.path.isdir(absolute_path):
            _hash = default_directory_hash
        # File
        elif os.path.exists(absolute_path):
            # Open the file and hash it using Dropbox python helpers
            hasher = DropboxContentHasher()
            with open(absolute_path, 'rb') as f:
                while True:
                    chunk = f.read(1024)  # or whatever chunk size you want
                    if len(chunk) == 0:
                        break
                    hasher.update(chunk)
            _hash = hasher.hexdigest()
        logging.getLogger(__name__).debug("Successfully computed hash of file (%.3f): %s" % (time.perf_counter() - duration, absolute_path))
    except:
        logging.getLogger(__name__).exception("Error while hashing file %s" % absolute_path)
    return _hash
    
    
    
    
class DropboxContentHasher(object):
    """
    Computes a hash using the same algorithm that the Dropbox API uses for the
    the "content_hash" metadata field.

    The digest() method returns a raw binary representation of the hash.  The
    hexdigest() convenience method returns a hexadecimal-encoded version, which
    is what the "content_hash" metadata field uses.

    How to use it:

    .. code-block:: python 

        hasher = DropboxContentHasher()
        with open('some-file', 'rb') as f:
            while True:
                chunk = f.read(1024)  # or whatever chunk size you want
                if len(chunk) == 0:
                    break
                hasher.update(chunk)
        print(hasher.hexdigest())

    """

    BLOCK_SIZE = 4 * 1024 * 1024

    def __init__(self):
        self._overall_hasher = hashlib.sha256()
        self._block_hasher = hashlib.sha256()
        self._block_pos = 0

        self.digest_size = self._overall_hasher.digest_size
        # hashlib classes also define 'block_size', but I don't know how people use that value

    def update(self, new_data):
        if self._overall_hasher is None:
            raise AssertionError(
                "can't use this object anymore; you already called digest()")

        # assert isinstance(new_data, six.binary_type), (
        #     "Expecting a byte string, got {!r}".format(new_data))

        new_data_pos = 0
        while new_data_pos < len(new_data):
            if self._block_pos == self.BLOCK_SIZE:
                self._overall_hasher.update(self._block_hasher.digest())
                self._block_hasher = hashlib.sha256()
                self._block_pos = 0

            space_in_block = self.BLOCK_SIZE - self._block_pos
            part = new_data[new_data_pos:(new_data_pos+space_in_block)]
            self._block_hasher.update(part)

            self._block_pos += len(part)
            new_data_pos += len(part)

    def _finish(self):
        if self._overall_hasher is None:
            raise AssertionError(
                "can't use this object anymore; you already called digest() or hexdigest()")

        if self._block_pos > 0:
            self._overall_hasher.update(self._block_hasher.digest())
            self._block_hasher = None
        h = self._overall_hasher
        self._overall_hasher = None  # Make sure we can't use this object anymore.
        return h

    def digest(self):
        return self._finish().digest()

    def hexdigest(self):
        return self._finish().hexdigest()

    def copy(self):
        c = DropboxContentHasher.__new__(DropboxContentHasher)
        c._overall_hasher = self._overall_hasher.copy()
        c._block_hasher = self._block_hasher.copy()
        c._block_pos = self._block_pos
        return c
    
    
            






#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
# Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: revised_watchdog.events
:synopsis: File system events and event handlers.
:author: yesudeep@google.com (Yesudeep Mangalapilly)
:author: Clément Warneys <clement.warneys@gmail.com>

This module is overloading the original :py:mod:`watchdog.events` module 
by revising and completing it. Please read original **watchdog** project 
documentation for more information: https://github.com/gorakhargosh/watchdog

This module imports some definitions of watchdog.events and keeps them unchanged:

* :py:class:`FileModifiedEvent`
* :py:class:`DirModifiedEvent`
* :py:class:`FileSystemEvent`
* :py:class:`FileSystemEventHandler`
* :py:data:`EVENT_TYPE_MOVED`
* :py:data:`EVENT_TYPE_CREATED`
* :py:data:`EVENT_TYPE_DELETED`

It adds the following definitions, in order to add some granularity in the
:py:class:`watchdog.events.ModifiedEvent` definition, thus differentiating content modification
from only metadata (access date, owner, etc.) modification:

* :py:class:`MetaFileModifiedEvent`
* :py:class:`TrueFileModifiedEvent`
* :py:class:`MetaDirModifiedEvent`
* :py:class:`TrueDirModifiedEvent`
* :py:data:`EVENT_TYPE_C_MODIFIED`
* :py:data:`EVENT_TYPE_M_MODIFIED`

Finally, it overloads the FileSystemEventHandler class, in order 
to manage the new granularity of modified events:

* :py:class:`FileSystemEventHandler`

"""

from watchdog.events import (
    FileModifiedEvent, 
    DirModifiedEvent, 
    FileSystemEvent,
    FileSystemEventHandler, 
    EVENT_TYPE_MOVED, 
    EVENT_TYPE_CREATED, 
    EVENT_TYPE_DELETED
    )

EVENT_TYPE_C_MODIFIED = 'modified' # content-only modification
EVENT_TYPE_M_MODIFIED = 'metadata' # metadata-only modification

class MetaFileModifiedEvent(FileModifiedEvent):
    """File system event representing metadata file modification on the file system."""

    event_type = EVENT_TYPE_M_MODIFIED

    def __init__(self, src_path):
        super(MetaFileModifiedEvent, self).__init__(src_path)

class TrueFileModifiedEvent(FileModifiedEvent):
    """File system event representing true file content modification on the file system."""

    event_type = EVENT_TYPE_C_MODIFIED

    def __init__(self, src_path):
        super(TrueFileModifiedEvent, self).__init__(src_path)

class MetaDirModifiedEvent(DirModifiedEvent):
    """File system event representing metadata directory modification on the file system."""

    event_type = EVENT_TYPE_M_MODIFIED

    def __init__(self, src_path):
        super(MetaDirModifiedEvent, self).__init__(src_path)

class TrueDirModifiedEvent(DirModifiedEvent):
    """File system event representing true directory content modification on the file system."""

    event_type = EVENT_TYPE_C_MODIFIED

    def __init__(self, src_path):
        super(TrueDirModifiedEvent, self).__init__(src_path)

class FileSystemEventHandler(FileSystemEventHandler):
    """
    Base file system event handler that you can override methods from.
    With modified dispatch method, added :py:meth:`on_data_modified` 
    and :py:meth:`on_meta_modified` methods, thus covering specific 
    needs of lazydog.
    """

    def dispatch(self, event):
        """
        Dispatches events to the appropriate methods.

        :param event:
            The event object representing the file system event.
        :type event:
            :py:class:`~watchdog.events.FileSystemEvent`
        """
        self.on_any_event(event)
        _method_map = {
            EVENT_TYPE_M_MODIFIED: self.on_meta_modified,
            EVENT_TYPE_C_MODIFIED: self.on_data_modified,
            EVENT_TYPE_MOVED: self.on_moved,
            EVENT_TYPE_CREATED: self.on_created,
            EVENT_TYPE_DELETED: self.on_deleted,
        }
        event_type = event.event_type
        _method_map[event_type](event)


    def on_data_modified(self, event):
        """Called when a file or directory true content is modified.

        :param event:
            Event representing file or directory modification.
        :type event:
            :py:class:`DirModifiedEvent` or :py:class:`FileModifiedEvent`
        """
        self.on_modified(event)
        
    def on_meta_modified(self, event):
        """Called when a file or directory metadata is modified.

        :param event:
            Event representing file or directory modification.
        :type event:
            :py:class:`DirModifiedEvent` or :py:class:`FileModifiedEvent`
        """
        self.on_modified(event)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: lazydog.queues
:synopsis: Bufferizes lazydog events pending for a possible \
aggregation with other simultaneous events.
:author: Clément Warneys <clement.warneys@gmail.com>

"""

from lazydog.states import LocalState
from lazydog.events import LazydogEvent
from lazydog.revised_watchdog.events import FileSystemEventHandler


class DatedlocaleventQueue(FileSystemEventHandler):
    """
    Basically accumulates all the events emited by a watchdog oberver.
    It inherits from :py:class:`~revised_watchdog.events.FileSystemEventHandler`, so it 
    is compatible with watchdog oberver. The :py:meth:`on_any_event` catches the 
    low-level event and adds them to the queue, after transorming them 
    to :py:class:`~lazydog.events.LazydogEvent`, which will further allow them
    to be post-treated by a :py:class:`~lazydog.handlers.HighlevelEventHandler`.

    The :py:class:`~lazydog.queues.DatedlocaleventQueue` has to be initialized 
    with a :py:class:`~lazydog.states.LocalState` object.
    """

    def __init__(self, local_states:LocalState):
        self.events_list = []
        self.events_list.clear()
        self.local_states = local_states
        super(DatedlocaleventQueue, self).__init__()

    def on_any_event(self, event):
        """
        Catch-all event handler.

        :param event:
            The event object representing the file system event.
        :type event:
            :py:class:`watchdog.events.FileSystemEvent`
        """
        super(DatedlocaleventQueue, self).on_any_event(event)
        self.events_list.append(LazydogEvent(event, self.local_states))

    def next(self):
        """
        Provides with the oldest event that has been queued, removing it 
        from the queue in the same time.
        """
        return self.events_list.pop(0) if not self.is_empty() else None
            
    def size(self):
        """Returns an integer corresponding to the current size of the queue."""
        return len(self.events_list)

    def is_empty(self):
        """``True`` if the queue size is 0."""
        return self.size() == 0
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Created on 17 janv. 2018

@author: Clement Warneys <clement.warneys@gmail.com>
'''

from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

import io
import codecs
import os
import sys

import lazydog

here = os.path.abspath(os.path.dirname(__file__))

def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

long_description = 'Python module monitoring high-level file system events ' + \
                   'like Creation, Modification, Move, Copy, and Deletion of ' + \
                   'files and folders. Lazydog tries to aggregate low-level ' + \
                   'events between them in order to emit a minimum number of ' + \
                   'high-level events (actualy one event per user action). ' + \
                   'Lazydog uses python Watchdog module to detect low-level events.'

class PyTest(TestCommand):
    
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

setup(
    name='lazydog',
    version=lazydog.__version__,
    url='http://github.com/warniiiz/lazydog/',
    license='Apache Software License',
    author='Clément Warneys',
    author_email='clement.warneys@gmail.com',

    python_requires='>=3.5',
    tests_require=['pytest'],
    
    #===========================================================================
    # # How to know current version on your system?
    # import pip
    # for x in pip.get_installed_distributions(): print(x)
    #===========================================================================
    
    install_requires=['watchdog>=0.8.3',
                    ],
    cmdclass={'test': PyTest},
    description='User-level filesystem event observer',

    long_description=long_description,
    keywords='python watchdog inotify monitoring watcher observer file filesystem filesystem-events copy move create delete modify detect',

    project_urls={
        'Documentation': 'http://lazydog.readthedocs.io/',
        'Source': 'https://github.com/warniiiz/lazydog',
    },

    packages=['lazydog', 'lazydog.revised_watchdog', 'lazydog.revised_watchdog.observers'],
    include_package_data=True,
    platforms='any',
    test_suite='lazydog.test',
    classifiers = [
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Filesystems',
        'Topic :: Desktop Environment :: File Managers',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],

    entry_points = {
        'console_scripts': ['lazydog=lazydog.lazydog:main'],
    },
    extras_require={
        'testing': ['pytest'],
    }
)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This inner package is overloading the original **watchdog** package 
by revising and completing it, resolving the current situation
where the useful **watchdog** package is not maintained anymore...

Please read original **watchdog** project documentation for more information:
https://pypi.org/project/watchdog/

revised_watchdog.events
=======================

.. automodule:: lazydog.revised_watchdog.events
   :members:



revised_watchdog.observers.inotify
==================================

.. automodule:: lazydog.revised_watchdog.observers.inotify
   :members:

revised_watchdog.observers.inotify_c
====================================

.. automodule:: lazydog.revised_watchdog.observers.inotify_c
   :members:
   
revised_watchdog.observers.inotify_buffer
=========================================

.. automodule:: lazydog.revised_watchdog.observers.inotify_buffer
   :members:

"""